/**
 * # HTML compiler
 *
 * This is a compiler for HyperText Markup Language (HTML).
 * It parses/unparses Stencila Document Tree (SDT) nodes from/to HTML.
 *
 * ## Philosophy
 *
 * ### Don't worry about lossy parsing from HTML
 *
 * The aim of this compiler is not to parse any old HTML file from off the interwebs.
 *
 * ### Aim for lossless unparsing to HTML
 *
 * The aim of this compiler is to be able to publish Stencila documents as
 * completely as possible. One way we achieve through JSON-LD metadata.
 *
 * ### Use custom elements where necessary
 *
 * For SDT nodes that do not have a natural HTML counterpart, we use
 * [custom elements](https://developer.mozilla.org/en-US/docs/Web/Web_Components/Using_custom_elements).
 *
 * ### Generate pretty HTML
 *
 * Because it's easier to debug.
 *
 * ## Dependencies
 *
 * Some of the main external dependencies:
 *
 * - `jsdom` for it's DOM API convenience with built in typing support .
 *   e.g. `doc.querySelector('p')` returns a `HTMLParagraphElement`.
 *
 * - `hyperscript` for it's beautifully minimal API with built in typing support.
 *   e.g. `h('p')` returns a `HTMLParagraphElement`
 *
 * - `collapse-whitespace` to avoid extraneous string elements in parsed content.
 *
 * - `js-beautify` for pretty generated HTML.
 *
 * - `json5` for more human readable representation of arbitrary values.
 *
 * @module html
 */
/** A comment required for above to be included in docs.
 * See https://github.com/christopherthielen/typedoc-plugin-external-module-name/issues/300
 */
import * as stencila from '@stencila/schema';
import { VFile } from './vfile';
export declare const mediaTypes: string[];
/**
 * Parse a `VFile` with HTML contents to a `stencila.Node`.
 *
 * @param file The `VFile` to parse
 * @returns A promise that resolves to a `stencila.Node`
 */
export declare function parse(file: VFile): Promise<stencila.Node>;
/**
 * Unparse a `stencila.Node` to a `VFile` with HTML contents.
 *
 * @param node The `stencila.Node` to unparse. Will be mutated to an `Node`.
 * @returns A promise that resolves to a `VFile`
 */
export declare function unparse(node: stencila.Node): Promise<VFile>;
