"use strict";
/**
 * # HTML compiler
 *
 * This is a compiler for HyperText Markup Language (HTML).
 * It parses/unparses Stencila Document Tree (SDT) nodes from/to HTML.
 *
 * ## Philosophy
 *
 * ### Don't worry about lossy parsing from HTML
 *
 * The aim of this compiler is not to parse any old HTML file from off the interwebs.
 *
 * ### Aim for lossless unparsing to HTML
 *
 * The aim of this compiler is to be able to publish Stencila documents as
 * completely as possible. One way we achieve through JSON-LD metadata.
 *
 * ### Use custom elements where necessary
 *
 * For SDT nodes that do not have a natural HTML counterpart, we use
 * [custom elements](https://developer.mozilla.org/en-US/docs/Web/Web_Components/Using_custom_elements).
 *
 * ### Generate pretty HTML
 *
 * Because it's easier to debug.
 *
 * ## Dependencies
 *
 * Some of the main external dependencies:
 *
 * - `jsdom` for it's DOM API convenience with built in typing support .
 *   e.g. `doc.querySelector('p')` returns a `HTMLParagraphElement`.
 *
 * - `hyperscript` for it's beautifully minimal API with built in typing support.
 *   e.g. `h('p')` returns a `HTMLParagraphElement`
 *
 * - `collapse-whitespace` to avoid extraneous string elements in parsed content.
 *
 * - `js-beautify` for pretty generated HTML.
 *
 * - `json5` for more human readable representation of arbitrary values.
 *
 * @module html
 */
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) if (e.indexOf(p[i]) < 0)
            t[p[i]] = s[p[i]];
    return t;
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
/** A comment required for above to be included in docs.
 * See https://github.com/christopherthielen/typedoc-plugin-external-module-name/issues/300
 */
const stencila = __importStar(require("@stencila/schema"));
const collapse_whitespace_1 = __importDefault(require("collapse-whitespace"));
const hyperscript_1 = __importDefault(require("hyperscript"));
// @ts-ignore
const js_beautify_1 = require("js-beautify");
const jsdom_1 = __importDefault(require("jsdom"));
const json5_1 = __importDefault(require("json5"));
const stencila_css_template_1 = require("./templates/stencila-css-template");
const vfile_1 = require("./vfile");
const document = new jsdom_1.default.JSDOM().window.document;
exports.mediaTypes = ['text/html'];
/**
 * Parse a `VFile` with HTML contents to a `stencila.Node`.
 *
 * @param file The `VFile` to parse
 * @returns A promise that resolves to a `stencila.Node`
 */
async function parse(file) {
    const html = await vfile_1.dump(file);
    const dom = new jsdom_1.default.JSDOM(html);
    const document = dom.window.document;
    collapse_whitespace_1.default(document);
    return parseNode(document);
}
exports.parse = parse;
/**
 * Unparse a `stencila.Node` to a `VFile` with HTML contents.
 *
 * @param node The `stencila.Node` to unparse. Will be mutated to an `Node`.
 * @returns A promise that resolves to a `VFile`
 */
async function unparse(node) {
    const dom = unparseNode(node);
    const beautifulHtml = js_beautify_1.html(dom.outerHTML, {
        indent_size: 2,
        indent_inner_html: true,
        wrap_line_length: 100,
        preserve_newlines: false // Preserve existing line-breaks
    });
    return vfile_1.load(beautifulHtml);
}
exports.unparse = unparse;
function parseNode(node) {
    const name = node.nodeName.toLowerCase();
    switch (name) {
        case '#document':
            return parseDocument(node);
        case 'p':
            return parseParagraph(node);
        case 'blockquote':
            return parseBlockquote(node);
        case 'pre':
            if (node.firstChild && node.firstChild.nodeName === 'CODE') {
                return parseCodeBlock(node);
            }
            break;
        case 'ul':
            return parseList(node);
        case 'ol':
            return parseList(node);
        case 'table':
            return parseTable(node);
        case 'hr':
            return parseHR(node);
        case 'em':
            return parseInlineElement(node, 'Emphasis');
        case 'strong':
            return parseInlineElement(node, 'Strong');
        case 'del':
            return parseInlineElement(node, 'Delete');
        case 'a':
            return parseLink(node);
        case 'q':
            return parseQuote(node);
        case 'code':
            return parseCode(node);
        case 'img':
            return parseImage(node);
        case 'stencila-null':
            return parseNull(node);
        case 'stencila-boolean':
            return parseBoolean(node);
        case 'stencila-number':
            return parseNumber(node);
        case 'stencila-array':
            return parseArray(node);
        case 'stencila-object':
            return parseObject(node);
        case 'stencila-thing':
            return parseThing(node);
        case '#text':
            return parseText(node);
    }
    const match = name.match(/^h(\d)$/);
    if (match) {
        return parseHeading(node, parseInt(match[1], 10));
    }
    throw new Error(`No HTML parser for HTML element <${name}>`);
}
function unparseNode(node) {
    const type = stencila.type(node);
    switch (type) {
        case 'Article':
            return unparseArticle(node);
        case 'Heading':
            return unparseHeading(node);
        case 'Paragraph':
            return unparseParagraph(node);
        case 'QuoteBlock':
            return unparseQuoteBlock(node);
        case 'CodeBlock':
            return unparseCodeBlock(node);
        case 'List':
            return unparseList(node);
        case 'Table':
            return unparseTable(node);
        case 'ThematicBreak':
            return unparseThematicBreak(node);
        case 'Emphasis':
            return unparseInlineThing(node, 'em');
        case 'Strong':
            return unparseInlineThing(node, 'strong');
        case 'Delete':
            return unparseInlineThing(node, 'del');
        case 'Link':
            return unparseLink(node);
        case 'Quote':
            return unparseQuote(node);
        case 'Code':
            return unparseCode(node);
        case 'ImageObject':
            return unparseImageObject(node);
        case 'null':
            return unparseNull(node);
        case 'boolean':
            return unparseBoolean(node);
        case 'number':
            return unparseNumber(node);
        case 'string':
            return unparseString(node);
        case 'array':
            return unparseArray(node);
        case 'object':
            return unparseObject(node);
        default:
            return unparseThing(node);
    }
}
function parseBlockChildNodes(node) {
    return Array.from(node.childNodes).map(child => parseNode(child));
}
function parseInlineChildNodes(node) {
    return Array.from(node.childNodes).map(child => parseNode(child));
}
/**
 * Parse a `#document` node to a `stencila.CreativeWork`.
 */
function parseDocument(doc) {
    const head = doc.querySelector('head');
    if (!head)
        throw new Error('Document does not have a <head>!');
    const body = doc.querySelector('body');
    if (!body)
        throw new Error('Document does not have a <body>!');
    const jsonld = head.querySelector('script[type="application/ld+json"]');
    const metadata = jsonld ? JSON.parse(jsonld.innerHTML || '{}') : {};
    delete metadata['@context'];
    // TODO: Allow for the different types of creative work based on type in
    // the jsonld
    return Object.assign({ type: 'Article' }, metadata, { content: parseBlockChildNodes(body) });
}
/**
 * Generate a `<html>` element with supplied title, metadata, body content, and
 * optionally custom CSS to style the document with.
 */
function generateHtmlElement(title = 'Untitled', metadata = {}, body = [], style = stencila_css_template_1.stencilaCSS) {
    // prettier-ignore
    return hyperscript_1.default('html', hyperscript_1.default('head', hyperscript_1.default('title', title), hyperscript_1.default('meta', { charset: 'utf-8' }), hyperscript_1.default('script', { type: 'application/ld+json' }, JSON.stringify(Object.assign({ '@context': 'http://stencila.github.io/schema/stencila.jsonld' }, metadata))), hyperscript_1.default('style', { innerHTML: style })), hyperscript_1.default('body', body));
}
/**
 * Unparse a `stencila.Article` to a `#document` node.
 */
function unparseArticle(article) {
    const { type, title, content } = article, rest = __rest(article, ["type", "title", "content"]);
    const metadata = Object.assign({ type, title }, rest);
    const body = content ? content.map(unparseNode) : [];
    return generateHtmlElement(title, metadata, body);
}
/**
 * Parse a `<h1>` etc element to a `stencila.Heading`.
 */
function parseHeading(heading, depth) {
    return { type: 'Heading', depth, content: parseInlineChildNodes(heading) };
}
/**
 * Unparse a `stencila.Heading` to a `<h1>` etc element.
 */
function unparseHeading(heading) {
    return hyperscript_1.default(`h${heading.depth}`, heading.content.map(unparseNode));
}
/**
 * Parse a `<p>` element to a `stencila.Paragraph`.
 */
function parseParagraph(para) {
    return { type: 'Paragraph', content: parseInlineChildNodes(para) };
}
/**
 * Unparse a `stencila.Paragraph` to a `<p>` element.
 */
function unparseParagraph(para) {
    return hyperscript_1.default('p', para.content.map(unparseNode));
}
/**
 * Parse a `<blockquote>` element to a `stencila.QuoteBlock`.
 */
function parseBlockquote(elem) {
    const quoteBlock = {
        type: 'QuoteBlock',
        content: parseBlockChildNodes(elem)
    };
    const cite = elem.getAttribute('cite');
    if (cite)
        quoteBlock.citation = cite;
    return quoteBlock;
}
/**
 * Unparse a `stencila.QuoteBlock` to a `<blockquote>` element.
 */
function unparseQuoteBlock(block) {
    return hyperscript_1.default('blockquote', { cite: block.citation }, block.content.map(unparseNode));
}
/**
 * Parse a `<pre><code class="language-xxx">` element to a `stencila.CodeBlock`.
 */
function parseCodeBlock(elem) {
    const code = elem.querySelector('code');
    if (!code)
        throw new Error('Woaah, this should never happen!');
    const { language, value } = parseCode(code);
    return {
        type: 'CodeBlock',
        language,
        value
    };
}
/**
 * Unparse a `stencila.CodeBlock` to a `<pre><code class="language-xxx">` element.
 */
function unparseCodeBlock(block) {
    const code = unparseCode(block);
    return hyperscript_1.default('pre', code);
}
/**
 * Parse a `<ul>` or `<ol>` element to a `stencila.List`.
 */
function parseList(list) {
    const order = list.tagName === 'UL' ? 'unordered' : 'ascending';
    return {
        type: 'List',
        order,
        items: Array.from(list.childNodes).map(
        // TODO: Currently assumes only one element per <li>
        (li) => li.firstChild ? parseNode(li.firstChild) : null)
    };
}
/**
 * Unparse a `stencila.List` to a `<ul>` or `<ol>` element.
 */
function unparseList(list) {
    return hyperscript_1.default(list.order === 'unordered' ? 'ul' : 'ol', list.items.map((item) => hyperscript_1.default('li', unparseNode(item))));
}
/**
 * Parse a `<table>` element to a `stencila.Table`.
 */
function parseTable(table) {
    return {
        type: 'Table',
        rows: Array.from(table.querySelectorAll('tr')).map((row) => {
            return {
                type: 'TableRow',
                cells: Array.from(row.querySelectorAll('td')).map((cell) => {
                    return {
                        type: 'TableCell',
                        content: parseInlineChildNodes(cell)
                    };
                })
            };
        })
    };
}
/**
 * Unparse a `stencila.Table` to a `<table>` element.
 */
function unparseTable(table) {
    // prettier-ignore
    return hyperscript_1.default('table', hyperscript_1.default('tbody', table.rows.map((row) => {
        return hyperscript_1.default('tr', row.cells.map((cell) => {
            return hyperscript_1.default('td', cell.content.map(unparseNode));
        }));
    })));
}
/**
 * Parse a `<hr>` element to a `stencila.ThematicBreak`.
 */
function parseHR(hr) {
    return { type: 'ThematicBreak' };
}
/**
 * Unparse a `stencila.ThematicBreak` to a `<hr>` element.
 */
function unparseThematicBreak(tb) {
    return hyperscript_1.default('hr');
}
/**
 * Parse an inline element e.g `<em>` to a inline `Thing` e.g. `Emphasis`.
 */
function parseInlineElement(elem, type) {
    return { type, content: parseInlineChildNodes(elem) };
}
/**
 * Unparse an inline `Thing` to an inline element e.g. `<em>`.
 */
function unparseInlineThing(node, tag) {
    node = node;
    // @ts-ignore
    return hyperscript_1.default(tag, node.content.map(unparseNode));
}
/**
 * Parse a `<a>` element to a `stencila.Link`.
 */
function parseLink(elem) {
    const href = elem.getAttribute('href') || '#';
    return { type: 'Link', target: href, content: parseInlineChildNodes(elem) };
}
/**
 * Unparse a `stencila.Link` to a `<a>` element.
 */
function unparseLink(link) {
    return hyperscript_1.default('a', { href: link.target }, link.content.map(unparseNode));
}
/**
 * Parse a `<q>` element to a `stencila.Quote`.
 */
function parseQuote(elem) {
    const quote = { type: 'Quote', content: [elem.innerHTML] };
    const cite = elem.getAttribute('cite');
    if (cite)
        quote.citation = cite;
    return quote;
}
/**
 * Unparse a `stencila.Quote` to a `<q>` element.
 */
function unparseQuote(quote) {
    return hyperscript_1.default('q', { cite: quote.citation }, quote.content);
}
/**
 * Parse a `<code>` element to a `stencila.Code`.
 */
function parseCode(elem) {
    const code = { type: 'Code', value: elem.textContent || '' };
    const clas = elem.getAttribute('class');
    if (clas) {
        const match = clas.match(/^language-(\w+)$/);
        if (match) {
            code.language = match[1];
        }
    }
    return code;
}
/**
 * Unparse a `stencila.Code` to a `<code>` element.
 */
function unparseCode(code) {
    const clas = code.language ? `language-${code.language}` : undefined;
    return hyperscript_1.default('code', { class: clas, innerHTML: code.value });
}
/**
 * Parse a HTML `<img>` element to a Stencila `ImageObject`.
 */
function parseImage(elem) {
    const image = {
        type: 'ImageObject',
        contentUrl: elem.src
    };
    if (elem.title)
        image.title = elem.title;
    if (elem.alt)
        image.text = elem.alt;
    return image;
}
/**
 * Unparse a Stencila `ImageObject` to a HTML `<img>` element.
 */
function unparseImageObject(image) {
    return hyperscript_1.default('img', {
        src: image.contentUrl,
        title: image.title,
        alt: image.text
    });
}
/**
 * Parse a `<stencila-null>` element to a `null`.
 */
function parseNull(elem) {
    return null;
}
/**
 * Unparse a `null` to a `<stencila-null>` element.
 */
function unparseNull(value) {
    return hyperscript_1.default('stencila-null', 'null');
}
/**
 * Parse a `<stencila-boolean>` element to a `boolean`.
 */
function parseBoolean(elem) {
    return elem.innerHTML === 'true' ? true : false;
}
/**
 * Unparse a `boolean` to a `<stencila-boolean>` element.
 */
function unparseBoolean(value) {
    return hyperscript_1.default('stencila-boolean', value === true ? 'true' : 'false');
}
/**
 * Parse a `<stencila-number>` element to a `number`.
 */
function parseNumber(elem) {
    return parseFloat(elem.innerHTML || '0');
}
/**
 * Unparse a `number` to a `<stencila-number>` element.
 */
function unparseNumber(value) {
    return hyperscript_1.default('stencila-number', value.toString());
}
/**
 * Parse a `<stencila-array>` element to a `array`.
 */
function parseArray(elem) {
    return json5_1.default.parse(elem.innerHTML || '[]');
}
/**
 * Unparse a `array` to a `<stencila-array>` element.
 */
function unparseArray(value) {
    return hyperscript_1.default('stencila-array', json5_1.default.stringify(value));
}
/**
 * Parse a `<stencila-object>` element to a `object`.
 */
function parseObject(elem) {
    return json5_1.default.parse(elem.innerHTML || '{}');
}
/**
 * Unparse a `object` to a `<stencila-object>` element.
 */
function unparseObject(value) {
    return hyperscript_1.default('stencila-object', json5_1.default.stringify(value));
}
/**
 * Parse a `<stencila-thing>` element to a `Thing`.
 */
function parseThing(elem) {
    return json5_1.default.parse(elem.innerHTML || '{}');
}
/**
 * Unparse a `Thing` to a `<stencila-thing>` element.
 */
function unparseThing(thing) {
    return hyperscript_1.default('stencila-thing', json5_1.default.stringify(thing));
}
/**
 * Parse a `#text` node to a `string`.
 */
function parseText(text) {
    return text.data;
}
/**
 * Unparse a `string` to a `#text` node.
 */
function unparseString(value) {
    return document.createTextNode(value);
}
