"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const mime_1 = __importDefault(require("mime"));
const path_1 = __importDefault(require("path"));
const csv = __importStar(require("./csv"));
const docx = __importStar(require("./docx"));
const gdoc = __importStar(require("./gdoc"));
const html = __importStar(require("./html"));
const jats = __importStar(require("./jats"));
const json = __importStar(require("./json"));
const json5 = __importStar(require("./json5"));
const latex = __importStar(require("./latex"));
const md = __importStar(require("./md"));
const ods = __importStar(require("./ods"));
const odt = __importStar(require("./odt"));
const pandoc = __importStar(require("./pandoc"));
const pdf = __importStar(require("./pdf"));
const rpng = __importStar(require("./rpng"));
const tdp = __importStar(require("./tdp"));
const vfile = __importStar(require("./vfile"));
const xlsx = __importStar(require("./xlsx"));
const yaml = __importStar(require("./yaml"));
/**
 * A list of all compilers.
 *
 * Note that order is of importance for matching. More "generic"
 * formats should go last. See the `match` function.
 */
exports.compilerList = [
    // Tabular data, spreadsheets etc
    csv,
    ods,
    tdp,
    xlsx,
    // Articles, textual documents etc
    docx,
    gdoc,
    html,
    jats,
    latex,
    md,
    odt,
    pdf,
    // Images
    rpng,
    // Data interchange formats
    yaml,
    pandoc,
    json5,
    json
];
/**
 * Match the compiler based on file name, extension name, media type or by content sniffing.
 *
 * Iterates through the list of compilers and returns the first that matches based on any
 * of the above criteria.
 *
 * If the supplied format contains a forward slash then it is assumed to be a media type,
 * otherwise an extension name.
 *
 * @param content The content as a file path (e.g. `../folder/file.txt`) or raw content
 * @param format The format as a media type (e.g. `text/plain`) or extension name (e.g. `txt`)
 * @returns A promise that resolves to the `Compiler` to use
 */
async function match(content, format) {
    // Resolve variables used to match a compiler...
    let fileName;
    let extName;
    let mediaType;
    // If the content is a path then begin with derived values
    if (content && vfile.isPath(content)) {
        fileName = path_1.default.basename(content);
        extName = path_1.default.extname(content).slice(1);
        mediaType = mime_1.default.getType(content) || undefined;
    }
    // But override with supplied format (if any) assuming that
    // media types always have a forward slash and extension names
    // never do.
    if (format) {
        if (format.includes('/'))
            mediaType = format;
        else {
            extName = format;
            mediaType = mime_1.default.getType(extName) || undefined;
        }
    }
    for (let compiler of exports.compilerList) {
        if (fileName &&
            compiler.fileNames &&
            compiler.fileNames.includes(fileName)) {
            return compiler;
        }
        if (extName && compiler.extNames && compiler.extNames.includes(extName)) {
            return compiler;
        }
        if (mediaType &&
            compiler.mediaTypes &&
            compiler.mediaTypes.includes(mediaType)) {
            return compiler;
        }
        if (content && compiler.sniff && (await compiler.sniff(content))) {
            return compiler;
        }
    }
    let message = 'No compiler could be found';
    if (content)
        message += ` for content "${content}"`;
    if (format)
        message += ` for format "${format}"`;
    message += '.';
    throw new Error(message);
}
exports.match = match;
/**
 * Is the file path, or media type handled? (i.e. is there a compiler for it?)
 *
 * @param content The file path
 * @param format The media type
 */
async function handled(content, format) {
    try {
        await match(content, format);
        return true;
    }
    catch (error) {
        return false;
    }
}
exports.handled = handled;
/**
 * Parse a virtual file to a `stencila.Node`
 *
 * @param file The `VFile` to parse
 * @param content The file path
 * @param format The media type
 */
async function parse(file, content, format) {
    const compiler = await match(content, format);
    return compiler.parse(file);
}
exports.parse = parse;
/**
 * Unparse (i.e. serialize) a `stencila.Node` to a virtual file.
 *
 * @param node The node to unparse
 * @param filePath The file path to unparse the node to.
 *                 Only required for some compilers e.g. those unparsing to more than one file.
 * @param format The format to unparse the node as.
 *               If undefined then determined from filePath or file path.
 */
async function unparse(node, filePath, format) {
    const compiler = await match(filePath, format);
    return compiler.unparse(node, filePath);
}
exports.unparse = unparse;
/**
 * Load a `stencila.Node` from a string of content.
 *
 * @param content The content to load.
 * @param format The format to load the content as.
 */
async function load(content, format) {
    const file = vfile.load(content);
    return parse(file, undefined, format);
}
exports.load = load;
/**
 * Dump a `stencila.Node` to a string of content.
 *
 * @param node The node to dump.
 * @param format The format to dump the node as.
 */
async function dump(node, format) {
    const file = await unparse(node, undefined, format);
    return vfile.dump(file);
}
exports.dump = dump;
/**
 * Read a file to a `stencila.Node`.
 *
 * @param content The raw content or file path to read.
 *                Use `-` to read from standard input.
 * @param format The format to read the file as.
 *               If undefined then determined from content or file path.
 */
async function read(content, format) {
    let file = await vfile.read(content);
    return parse(file, content, format);
}
exports.read = read;
/**
 * Write a `stencila.Node` to a file.
 *
 * @param node The node to write.
 * @param filePath The file system path to write to.
 *                 Use `-` write to standard output.
 * @param format The format to write the node as.
 */
async function write(node, filePath, format) {
    let file = await unparse(node, filePath, format);
    await vfile.write(file, filePath);
    return file;
}
exports.write = write;
/**
 * Convert content from one format to another.
 *
 * @param input The input content (raw or file path).
 * @param outputPath The output file path.
 * @param options Conversion options e.g `from` and `to`: to specify the formats to convert from/to
 * @returns The converted content, or file path (for converters that only write to files).
 */
async function convert(input, outputPath, options = {}) {
    const inputFile = vfile.create(input);
    const node = await parse(inputFile, input, options.from);
    const outputFile = await unparse(node, outputPath, options.to);
    if (outputPath)
        await vfile.write(outputFile, outputPath);
    return outputFile.contents ? vfile.dump(outputFile) : outputFile.path;
}
exports.convert = convert;
