"use strict";
/**
 * # JSON5 compiler
 *
 * [JSON5](https://json5.org/) is "JSON for Humans":
 *
 * > The JSON5 Data Interchange Format (JSON5) is a superset of JSON that aims
 * > to alleviate some of the limitations of JSON by expanding its syntax to
 * > include some productions from ECMAScript 5.1.
 *
 * This compiler is primarily targeted at developers.
 * Given that it is more forgiving and it has less typing overhead,
 * JSON5 can be a little more convenient than JSON for quickly testing
 * unparsing of Stencila nodes e.g
 *
 * ```bash
 * stencila convert "{type:'ImageObject', contentUrl: 'https://example.org', text: 'alt', title: 'title'}" --from json5 --to html
 * ```
 *
 * Versus using JSON:
 *
 * ```bash
 * stencila convert '{"type":"ImageObject", "contentUrl": "https://example.org", "text": "alt", "title": "title"}' --from json --to html
 * ```
 *
 * @module json5
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const json5_1 = __importDefault(require("json5"));
const vfile_1 = require("./vfile");
/**
 * The media types that this compiler can parse/unparse.
 */
exports.mediaTypes = ['application/json5'];
// The above media type is registered in the `mime` module
// so there is no need to specify `extNames`
/**
 * Parse a `VFile` with JSON5 content to a Stencila `Node`.
 *
 * @param file The `VFile` to parse
 * @returns A promise that resolves to a Stencila `Node`
 */
async function parse(file) {
    return json5_1.default.parse(await vfile_1.dump(file));
}
exports.parse = parse;
/**
 * Unparse a Stencila `Node` to a `VFile` with JSON5 content.
 *
 * @param thing The Stencila `Node` to unparse
 * @returns A promise that resolves to a `VFile`
 */
async function unparse(node) {
    return vfile_1.load(json5_1.default.stringify(node, null, '  '));
}
exports.unparse = unparse;
