"use strict";
/**
 * # Markdown compiler
 *
 * These functions transform nodes from a [Markdown Abstract Syntax Tree](https://github.com/syntax-tree/mdast) to
 * nodes in a [Stencila Document Tree](https://github.com/stencila/schema).
 *
 * @module md
 */
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
/** A comment required for above to be included in docs.
 * See https://github.com/christopherthielen/typedoc-plugin-external-module-name/issues/300
 */
const stencila = __importStar(require("@stencila/schema"));
const yaml = __importStar(require("js-yaml"));
const json5_1 = __importDefault(require("json5"));
// @ts-ignore
const mdast_util_compact_1 = __importDefault(require("mdast-util-compact"));
// @ts-ignore
const remark_frontmatter_1 = __importDefault(require("remark-frontmatter"));
// @ts-ignore
const remark_generic_extensions_1 = __importDefault(require("remark-generic-extensions"));
// @ts-ignore
const remark_parse_1 = __importDefault(require("remark-parse"));
// @ts-ignore
const remark_stringify_1 = __importDefault(require("remark-stringify"));
const unified_1 = __importDefault(require("unified"));
// @ts-ignore
const unist_util_filter_1 = __importDefault(require("unist-util-filter"));
// @ts-ignore
const unist_util_map_1 = __importDefault(require("unist-util-map"));
const util_1 = require("./util");
const vfile_1 = require("./vfile");
exports.mediaTypes = ['text/markdown', 'text/x-markdown'];
/**
 * Options for `remark-frontmatter` parser and stringifier
 *
 * @see https://github.com/remarkjs/remark-frontmatter#matter
 */
const FRONTMATTER_OPTIONS = [{ type: 'yaml', marker: '-' }];
/**
 * Registered generic extensions.
 *
 * @see Extension
 */
const GENERIC_EXTENSIONS = [
    'quote',
    'null',
    'true',
    'false',
    'boolean',
    'number',
    'array',
    'object'
];
/**
 * Parse a `VFile` with Markdown contents to a `stencila.Node`.
 *
 * @param file The `VFile` to parse
 * @returns A promise that resolves to a `stencila.Node`
 */
async function parse(file) {
    const extensionHandlers = {};
    for (let ext of GENERIC_EXTENSIONS) {
        extensionHandlers[ext] = { replace: parseExtension };
    }
    const mdast = unified_1.default()
        .use(remark_parse_1.default, {
        commonmark: true
    })
        .use(remark_frontmatter_1.default, FRONTMATTER_OPTIONS)
        .use(remark_generic_extensions_1.default, { elements: extensionHandlers })
        .parse(file);
    mdast_util_compact_1.default(mdast, true);
    return parseNode(mdast);
}
exports.parse = parse;
/**
 * Unparse a `stencila.Node` to a `VFile` with Markdown contents.
 *
 * @param thing The `stencila.Node` to unparse
 * @returns A promise that resolves to a `VFile`
 */
async function unparse(node) {
    let mdast = unist_util_filter_1.default(unparseNode(node), (node) => typeof node !== 'undefined');
    mdast = stringifyExtensions(mdast);
    const md = unified_1.default()
        .use(remark_stringify_1.default)
        .use(remark_frontmatter_1.default, FRONTMATTER_OPTIONS)
        .stringify(mdast);
    return vfile_1.load(md);
}
exports.unparse = unparse;
function parseNode(node) {
    const type = node.type;
    switch (type) {
        case 'root':
            return parseRoot(node);
        case 'heading':
            return parseHeading(node);
        case 'paragraph':
            return parseParagraph(node);
        case 'blockquote':
            return parseBlockquote(node);
        case 'code':
            return parseCodeblock(node);
        case 'list':
            return parseList(node);
        case 'table':
            return parseTable(node);
        case 'thematicBreak':
            return parseThematicBreak(node);
        case 'link':
            return parseLink(node);
        case 'emphasis':
            return parseEmphasis(node);
        case 'strong':
            return parseStrong(node);
        case 'delete':
            return parseDelete(node);
        case 'inlineCode':
            return parseInlineCode(node);
        case 'image':
            return parseImage(node);
        case 'text':
            return parseText(node);
        case 'inline-extension':
            const ext = node;
            switch (ext.name) {
                case 'quote':
                    return parseQuote(ext);
                case 'null':
                    return parseNull(ext);
                case 'boolean':
                case 'true':
                case 'false':
                    return parseBoolean(ext);
                case 'number':
                    return parseNumber(ext);
                case 'array':
                    return parseArray(ext);
                case 'object':
                    return parseObject(ext);
                default:
                    if (ext.name) {
                        throw new Error(`Unhandled generic extension "${ext.name}"`);
                    }
                    else {
                        throw new Error(`Unregistered generic extension "${node.data && node.data.hName}"`);
                    }
            }
        default:
            throw new Error(`No Markdown parser for MDAST node type "${type}"`);
    }
}
function unparseNode(node) {
    const type = stencila.type(node);
    switch (type) {
        case 'Article':
            return unparseArticle(node);
        case 'Heading':
            return unparseHeading(node);
        case 'Paragraph':
            return unparseParagraph(node);
        case 'QuoteBlock':
            return unparseQuoteBlock(node);
        case 'CodeBlock':
            return unparseCodeBlock(node);
        case 'List':
            return unparseList(node);
        case 'Table':
            return unparseTable(node);
        case 'ThematicBreak':
            return unparseThematicBreak(node);
        case 'Link':
            return unparseLink(node);
        case 'Emphasis':
            return unparseEmphasis(node);
        case 'Strong':
            return unparseStrong(node);
        case 'Delete':
            return unparseDelete(node);
        case 'Quote':
            return unparseQuote(node);
        case 'Code':
            return unparseCode(node);
        case 'ImageObject':
            return unparseImageObject(node);
        case 'string':
            return unparseString(node);
        case 'null':
            return unparseNull(node);
        case 'boolean':
            return unparseBoolean(node);
        case 'number':
            return unparseNumber(node);
        case 'array':
            return unparseArray(node);
        case 'object':
            return unparseObject(node);
        default:
            throw new Error(`No Markdown unparser for Stencila node type "${type}"`);
    }
}
function unparseContent(node) {
    return unparseNode(node);
}
function parsePhrasingContent(node) {
    return parseNode(node);
}
function unparseInlineContent(node) {
    return unparseNode(node);
}
function parseBlockContent(node) {
    return parseNode(node);
}
function unparseBlockContent(node) {
    return unparseNode(node);
}
/**
 * Parse a `MDAST.root` node to a `stencila.Article`
 *
 * If the root has a front matter node (defined using YAML), that
 * meta data is added to the top level of the document. Other
 * child nodes are added to the article's `content` property.
 *
 * @param root The MDAST root to parse
 */
function parseRoot(root) {
    const article = stencila.create('Article', {
        // TODO: the `create function should automatically add empty
        // array for array properties that are required
        authors: []
    }, 'mutate');
    const body = [];
    for (let child of root.children) {
        if (child.type === 'yaml') {
            const frontmatter = yaml.safeLoad(child.value);
            // TODO: check the key is a valid property of Article
            // and if it it isn't ignore it or throw an error
            // TODO: allow for mutation and aliases, potentially
            // adding a `stencila.set(article, key, value)` function.
            for (let [key, value] of Object.entries(frontmatter)) {
                // TODO: the above should allow removal of the ts-ignore
                // @ts-ignore
                article[key] = value;
            }
        }
        else {
            body.push(parseNode(child));
        }
    }
    article.content = body;
    // TODO: remove the following which mutates any YAML
    // meta data to conform to the schema when above TODO is added
    return util_1.coerce(article, 'Article');
}
/**
 * Unparse a `stencila.Article` to a `MDAST.Root`
 *
 * The article's `content` property becomes the root's `children`
 * and any other properties are serialized as YAML
 * front matter and prepended to the children.
 *
 * @param node The Stencila article to unparse
 */
function unparseArticle(article) {
    const root = {
        type: 'root',
        children: []
    };
    // Unparse the article body
    if (article.content) {
        root.children = article.content.map(unparseContent);
    }
    // Add other properties as frontmatter
    const frontmatter = {};
    for (let [key, value] of Object.entries(article)) {
        if (!['type', 'content'].includes(key)) {
            frontmatter[key] = value;
        }
    }
    if (Object.keys(frontmatter).length) {
        const yamlNode = {
            type: 'yaml',
            value: yaml.safeDump(frontmatter).trim()
        };
        root.children.unshift(yamlNode);
    }
    return root;
}
/**
 * Parse a `MDAST.Heading` to a `stencila.Heading`
 */
function parseHeading(heading) {
    return {
        type: 'Heading',
        depth: heading.depth,
        content: heading.children.map(parsePhrasingContent)
    };
}
/**
 * Unparse a `stencila.Heading` to a `MDAST.Heading`
 */
function unparseHeading(heading) {
    return {
        type: 'heading',
        depth: heading.depth,
        children: heading.content.map(unparseInlineContent)
    };
}
/**
 * Parse a `MDAST.Paragraph` to a `stencila.Paragraph`
 */
function parseParagraph(paragraph) {
    return {
        type: 'Paragraph',
        content: paragraph.children.map(parsePhrasingContent)
    };
}
/**
 * Unparse a `stencila.Paragraph` to a `MDAST.Paragraph`
 *
 * Returns `undefined` (i.e skip this node) if the paragraph
 * is empty (not content, or only whitespace)
 */
function unparseParagraph(paragraph) {
    const content = paragraph.content;
    if (content.length === 0 ||
        (content.length === 1 &&
            stencila.type(content[0]) === 'string' &&
            content[0].trim().length === 0)) {
        return undefined;
    }
    else {
        return {
            type: 'paragraph',
            children: content.map(unparseInlineContent)
        };
    }
}
/**
 * Parse a `MDAST.Blockquote` to a `stencila.QuoteBlock`
 */
function parseBlockquote(block) {
    return {
        type: 'QuoteBlock',
        content: block.children.map(parseBlockContent)
    };
}
/**
 * Unparse a `stencila.QuoteBlock` to a `MDAST.Blockquote`
 */
function unparseQuoteBlock(block) {
    return {
        type: 'blockquote',
        children: block.content.map(unparseBlockContent)
    };
}
/**
 * Parse a `MDAST.Code` to a `stencila.QuoteBlock`
 */
function parseCodeblock(block) {
    return {
        type: 'CodeBlock',
        language: block.lang,
        value: block.value
    };
}
/**
 * Unparse a `stencila.CodeBlock` to a `MDAST.Code`
 */
function unparseCodeBlock(block) {
    return {
        type: 'code',
        lang: block.language,
        value: block.value
    };
}
/**
 * Parse a `MDAST.List` to a `stencila.List`
 */
function parseList(list) {
    const items = [];
    for (let item of list.children) {
        // TODO: when there are more than one child then create a stencila.Block
        let node = parseNode(item.children[0]);
        // If the item has a check box then insert that as a boolean as the first
        // child of the first child
        if (item.checked === true || item.checked === false) {
            // @ts-ignore
            if (node.content)
                node.content = [item.checked, ...node.content];
        }
        items.push(node);
    }
    return {
        type: 'List',
        order: list.ordered ? 'ascending' : 'unordered',
        items
    };
}
/**
 * Unparse a `stencila.List` to a `MDAST.List`
 */
function unparseList(list) {
    return {
        type: 'list',
        ordered: list.order === 'ascending',
        children: list.items.map((item) => {
            // TODO: wrap anything that is not inline content into a block e.g. para
            const first = unparseNode(item);
            const children = [first];
            // Is this a checked item (ie. a paragraph starting with a boolean)?
            let checked = undefined;
            if (first.type === 'paragraph') {
                // @ts-ignore
                if (first.children[0].type === 'inline-extension') {
                    if (['true', 'false'].includes(first.children[0].name)) {
                        // Apply the boolean name to this list item
                        checked = first.children[0].name === 'true';
                        // Remove the boolean checkbox from the paragraph (since remark stringify does that)
                        first.children = first.children.slice(1);
                    }
                }
            }
            return {
                type: 'listItem',
                checked,
                children
            };
        })
    };
}
/**
 * Parse a `MDAST.Table` to a `stencila.Table`
 */
function parseTable(table) {
    return {
        type: 'Table',
        rows: table.children.map((row) => {
            return {
                type: 'TableRow',
                cells: row.children.map((cell) => {
                    return {
                        type: 'TableCell',
                        content: cell.children.map(parsePhrasingContent)
                    };
                })
            };
        })
    };
}
/**
 * Unparse a `stencila.Table` to a `MDAST.Table`
 */
function unparseTable(table) {
    return {
        type: 'table',
        children: table.rows.map((row) => {
            return {
                type: 'tableRow',
                children: row.cells.map((cell) => {
                    return {
                        type: 'tableCell',
                        children: cell.content.map(unparseInlineContent)
                    };
                })
            };
        })
    };
}
/**
 * Parse a `MDAST.ThematicBreak` to a `stencila.ThematicBreak`
 */
function parseThematicBreak(tbreak) {
    return {
        type: 'ThematicBreak'
    };
}
/**
 * Unparse a `stencila.ThematicBreak` to a `MDAST.ThematicBreak`
 */
function unparseThematicBreak(tbreak) {
    return {
        type: 'thematicBreak'
    };
}
/**
 * Parse a `MDAST.Link` to a `stencila.Link`
 */
function parseLink(link) {
    return {
        type: 'Link',
        target: link.url,
        content: link.children.map(parsePhrasingContent)
    };
}
/**
 * Unparse a `stencila.Link` to a `MDAST.Link`
 */
function unparseLink(link) {
    return {
        type: 'link',
        url: link.target,
        children: link.content.map(node => unparseInlineContent(node))
    };
}
/**
 * Parse a `MDAST.Emphasis` to a `stencila.Emphasis`
 */
function parseEmphasis(emphasis) {
    return {
        type: 'Emphasis',
        content: emphasis.children.map(parsePhrasingContent)
    };
}
/**
 * Unparse a `stencila.Emphasis` to a `MDAST.Emphasis`
 */
function unparseEmphasis(emphasis) {
    return {
        type: 'emphasis',
        children: emphasis.content.map(unparseInlineContent)
    };
}
/**
 * Parse a `MDAST.Strong` to a `stencila.Strong`
 */
function parseStrong(strong) {
    return {
        type: 'Strong',
        content: strong.children.map(parsePhrasingContent)
    };
}
/**
 * Unparse a `stencila.Strong` to a `MDAST.Strong`
 */
function unparseStrong(strong) {
    return {
        type: 'strong',
        children: strong.content.map(unparseInlineContent)
    };
}
/**
 * Parse a `MDAST.Delete` to a `stencila.Delete`
 */
function parseDelete(delet) {
    return {
        type: 'Delete',
        content: delet.children.map(parsePhrasingContent)
    };
}
/**
 * Unparse a `stencila.Delete` to a `MDAST.Delete`
 */
function unparseDelete(delet) {
    return {
        type: 'delete',
        children: delet.content.map(unparseInlineContent)
    };
}
/**
 * Parse a `!quote` inline extension to a `Quote`.
 *
 * Valid quotes include:
 *
 *   - `!quote[Quoted content]`
 *   - `!quote[Quoted content with _emphasis_](https://example.org)`
 */
function parseQuote(ext) {
    const quote = {
        type: 'Quote',
        // TODO: possibly parse the ext.content as Markdown?
        content: ext.content ? [ext.content] : []
    };
    const cite = ext.argument;
    if (cite)
        quote.citation = cite;
    return quote;
}
/**
 * Unparse a `stencila.Quote` to a `!quote` inline extension
 */
function unparseQuote(quote) {
    return {
        type: 'inline-extension',
        name: 'quote',
        // TODO: Handle cases where content is more than one string
        content: quote.content[0],
        argument: quote.citation
    };
}
/**
 * Parse a `MDAST.InlineCode` to a `stencila.Code`
 */
function parseInlineCode(inlineCode) {
    return {
        type: 'Code',
        value: inlineCode.value
    };
}
/**
 * Unparse a `stencila.Code` to a `MDAST.InlineCode`
 */
function unparseCode(code) {
    return {
        type: 'inlineCode',
        value: code.value
    };
}
/**
 * Parse a `MDAST.Image` to a `stencila.ImageObject`
 */
function parseImage(image) {
    const imageObject = {
        type: 'ImageObject',
        contentUrl: image.url
    };
    if (image.title)
        imageObject.title = image.title;
    if (image.alt)
        imageObject.text = image.alt;
    return imageObject;
}
/**
 * Unparse a `stencila.ImageObject` to a `MDAST.Image`
 */
function unparseImageObject(imageObject) {
    const image = {
        type: 'image',
        url: imageObject.contentUrl || ''
    };
    if (imageObject.title)
        image.title = imageObject.title;
    if (imageObject.text)
        image.alt = imageObject.text;
    return image;
}
/**
 * Parse a `MDAST.Text` to a `string`
 */
function parseText(text) {
    return text.value;
}
/**
 * Unparse a `string` to a `MDAST.Text`
 */
function unparseString(value) {
    return { type: 'text', value };
}
/**
 * Parse a `!null` inline extension to `null`
 */
function parseNull(ext) {
    return null;
}
/**
 * Unparse `null` to a `!null` inline extension
 */
function unparseNull(value) {
    return { type: 'inline-extension', name: 'null' };
}
/**
 * Parse a `!true`, `!false`, `!boolean` inline extension to a `boolean`
 *
 * Valid booleans include (the first two are the preferred and the default,
 * the last should be avoided):
 *
 *   - `!true` or `!false`
 *   - `!boolean(true)` and `!boolean(1)`
 *   - `!boolean(false)` and `!boolean(0)`
 *   - `!boolean` (parsed to `true`)
 *   - `!boolean[true]` and `!boolean[1]` etc
 */
function parseBoolean(ext) {
    switch (ext.name) {
        case 'true':
            return true;
        case 'false':
            return false;
        default:
            const value = ext.argument || ext.content || 'true';
            return value === 'true' || value === '1' ? true : false;
    }
}
/**
 * Unparse a `boolean` to a `!true` or `!false`.
 */
function unparseBoolean(value) {
    return { type: 'inline-extension', name: value ? 'true' : 'false' };
}
/**
 * Parse a `!number` inline extension to a `number`.
 *
 * Valid numbers include (the first is the preferred and the default,
 * the last should be avoided):
 *
 *   - `!number(3.14)`
 *   - `!number` (parsed to `0`)
 *   - `!number[3.14]`
 */
function parseNumber(ext) {
    return parseFloat(ext.argument || ext.content || '0');
}
/**
 * Unparse a `number` to a `!number` inline extension
 */
function unparseNumber(value) {
    return {
        type: 'inline-extension',
        name: 'number',
        argument: value.toString()
    };
}
/**
 * Parse an `!array` inline extension to an `Array`.
 *
 * Valid arrays include (the first is the preferred and the default,
 * the last should be avoided):
 *
 *   - `!array(1, 2)`
 *   - `!array` (parsed to `[]`)
 *   - `!array[1, 2]`
 */
function parseArray(ext) {
    const items = ext.argument || ext.content || '';
    const array = json5_1.default.parse(`[${items}]`);
    return array;
}
/**
 * Unparse an `array` to a `!array` inline extension
 */
function unparseArray(value) {
    const argument = json5_1.default.stringify(value).slice(1, -1);
    return { type: 'inline-extension', name: 'array', argument };
}
/**
 * Parse an `!object` inline extension to an `Object`.
 *
 * Valid objects include (the first is the preferred and the default,
 * the last should be avoided):
 *
 *   - `!object("key":value, ...)` (comma separated pairs, values can be any JSON primitives)
 *   - `!object{key=string ...}` (space separated pairs; values can only be strings)
 *   - `!object` (parsed to `{}`)
 *   - `!object["key":"value", ...]`
 */
function parseObject(ext) {
    if (ext.properties) {
        // Extension properties always contain `className` and `id`, which may
        // be undefined, so drop them.
        const props = {};
        for (let [key, value] of Object.entries(ext.properties)) {
            // tslint:disable-next-line
            if (typeof value !== 'undefined')
                props[key] = value;
        }
        if (Object.keys(props).length > 0)
            return props;
    }
    return json5_1.default.parse(`{${ext.argument || ext.content}}`);
}
/**
 * Unparse an `object` to a `!object` inline extension
 */
function unparseObject(value) {
    const argument = json5_1.default.stringify(value).slice(1, -1);
    return { type: 'inline-extension', name: 'object', argument };
}
/**
 * Parse a generic extension into an MDAST node.
 */
function parseExtension(type, element) {
    return Object.assign({ type }, element);
}
/**
 * Unparse a generic extension node into a `MDAST.HTML` node.
 *
 * The `remark-generic-extensions` does not do this stringifying for us.
 * We transform each to a `MDAST.HTML` node so that no escaping of the value is done
 * while stringifying.
 */
function stringifyExtensions(tree) {
    return unist_util_map_1.default(tree, (node) => {
        if (node.type === 'inline-extension' || node.type === 'block-extension') {
            const props = Object.entries(node.properties || {})
                .map(([key, value]) => `${key}=${JSON.stringify(value)}`)
                .join(' ');
            let value;
            if (node.type === 'inline-extension') {
                value = `!${node.name}`;
                if (node.content)
                    value += `[${node.content}]`;
                if (node.argument)
                    value += `(${node.argument})`;
                if (node.properties)
                    value += `{${props}}`;
            }
            else {
                value = `${node.name}:`;
                if (node.argument)
                    value += `${node.argument}`;
                value += `\n:::\n[${node.content || ''}]\n:::\n`;
                if (node.properties)
                    value += `{${props}}`;
            }
            return { type: 'html', value };
        }
        return node;
    });
}
