"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) if (e.indexOf(p[i]) < 0)
            t[p[i]] = s[p[i]];
    return t;
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const stencila = __importStar(require("@stencila/schema"));
const child_process_1 = __importDefault(require("child_process"));
const boot_1 = require("./boot");
const Pandoc = __importStar(require("./pandoc-types"));
const rpng = __importStar(require("./rpng"));
const vfile_1 = require("./vfile");
var pandoc_types_1 = require("./pandoc-types");
exports.InputFormat = pandoc_types_1.InputFormat;
exports.OutputFormat = pandoc_types_1.OutputFormat;
// Although this compiler is usually used as a base for others (e.g `docx`),
// the following definitions allow Pandoc JSON to be parsed or unparsed
// directly
exports.mediaTypes = ['application/pandoc+json'];
exports.extNames = ['pandoc'];
/**
 * Parse a `VFile` to a `stencila.Node`.
 *
 * @param file The `VFile` to parse
 * @returns A promise that resolves to a `stencila.Node`
 */
async function parse(file, from = Pandoc.InputFormat.json, options = [], ensureFile = false) {
    const args = [`--from=${from}`, `--to=json`].concat(options);
    let content = file.contents;
    if (!content || ensureFile) {
        if (ensureFile && !file.path)
            throw new Error('Must supply a file');
        args.push(`${file.path}`);
    }
    const json = await run(content, args);
    const pdoc = JSON.parse(json);
    return parseDocument(pdoc);
}
exports.parse = parse;
/**
 * Unparse a `stencila.Node` to a `VFile`.
 *
 * @param node The `stencila.Node` to unparse
 * @param options Additional arguments to pass to Pandoc
 * @param ensureFile Ensure that the output is a real file (ie. not stdout?)
 * @returns A promise that resolves to a `VFile`
 */
async function unparse(node, filePath, to = Pandoc.OutputFormat.json, options = [], ensureFile = false) {
    const type = stencila.type(node);
    if (type !== 'Article') {
        throw new Error(`Unable to unparse top level Stencila node of type:\n "${type}"\n Expected node of type "Article".`);
    }
    unparsePromises = [];
    const pdoc = unparseArticle(node);
    await Promise.all(unparsePromises);
    const args = [`--from=json`, `--to=${to}`].concat(options);
    if ((filePath && filePath !== '-') || ensureFile) {
        let output;
        if (!filePath || filePath === '-') {
            // Create a new file path, which is returned as `vfile.path`
            output = stencila.type(node).toLowerCase() + '.' + to;
            filePath = output;
        }
        else
            output = filePath;
        args.push(`--output=${output}`);
    }
    const json = JSON.stringify(pdoc);
    const content = await run(json, args);
    // If content was outputted, then load that into a vfile,
    // otherwise the vfile simply has path to the file created
    if (content)
        return vfile_1.load(content);
    else
        return vfile_1.create(undefined, { path: filePath });
}
exports.unparse = unparse;
/**
 * Promises for resources (usually rPNGs) to
 * be generated during unparsing.
 *
 * This approach of having a global set of promises
 * is a bit hacky but allows us to keep the nested chain
 * of unparsing functions synchronous.
 */
let unparsePromises = [];
/**
 * Run the Pandoc binary
 */
function run(input, args) {
    args.push(`--data-dir=${boot_1.pandocDataDir}`);
    if (process.env.DEBUG) {
        console.log(`Running ${boot_1.pandocPath} with args:\n  ${args.join('\n  ')}`);
    }
    return new Promise((resolve, reject) => {
        const child = child_process_1.default.spawn(boot_1.pandocPath, args);
        let stdout = '';
        let stderr = '';
        child.stdout.on('data', data => {
            stdout += data;
        });
        child.stderr.on('data', data => {
            stderr += data;
        });
        child.on('close', () => {
            if (stderr) {
                stderr = `Pandoc error\n  message: ${stderr}  args:\n    ${args.join('\n    ')}\n`;
                reject(new Error(stderr));
            }
            else
                resolve(stdout);
        });
        child.on('error', err => {
            reject(err);
        });
        if (input && input.length) {
            child.stdin.write(input, err => {
                if (err)
                    return reject(err);
                child.stdin.end();
            });
        }
    });
}
/**
 * Parse a Pandoc `Document` to a Stencila `Article`.
 */
function parseDocument(pdoc) {
    const meta = parseMeta(pdoc.meta);
    // TODO: mutate metadata to conform to schema
    const content = parseBlocks(pdoc.blocks);
    return Object.assign({ type: 'Article', authors: [] }, meta, { content });
}
/**
 * Parse a Stencila `Article` to a Pandoc `Document`.
 */
function unparseArticle(article) {
    const { type, content } = article, rest = __rest(article, ["type", "content"]);
    const meta = unparseMeta(rest);
    // TODO: wrap nodes as necessary and avoid use of `as`
    const blocks = unparseBlocks(content);
    return {
        'pandoc-api-version': Pandoc.Version,
        meta,
        blocks
    };
}
/**
 * Parse a Pandoc `Meta` node to an `object`
 */
function parseMeta(meta) {
    return objectMap(meta, (key, value) => parseMetaValue(value));
}
exports.parseMeta = parseMeta;
/**
 * Unparse an `object` of metadata into a Pandoc `Meta` node
 */
function unparseMeta(obj) {
    return objectMap(obj, (key, value) => unparseMetaValue(value));
}
exports.unparseMeta = unparseMeta;
/**
 * Parse a Pandoc `MetaValue` to a Stencila `Node`
 */
function parseMetaValue(value) {
    switch (value.t) {
        case 'MetaBool':
            return value.c;
        case 'MetaString':
            if (value.c === '!!null')
                return null;
            if (value.c.slice(0, 9) === '!!number ') {
                return parseFloat(value.c.slice(9));
            }
            return value.c;
        case 'MetaList':
            return value.c.map(parseMetaValue);
        case 'MetaMap':
            return objectMap(value.c, (key, value) => parseMetaValue(value));
        case 'MetaInlines':
            return {
                type: 'Paragraph',
                content: parseInlines(value.c)
            };
        case 'MetaBlocks':
            return {
                // TODO: Currently there is no stencila.Division
                // so using QuoteBlock instead
                type: 'QuoteBlock',
                content: parseBlocks(value.c)
            };
    }
}
/**
 * Unparse a Stencila `Node` to a Pandoc `MetaValue`
 *
 * For `null` and `number`, use a YAML "tags" syntax e.g. `!!null`
 * encoded into a Pandoc `MetaString`.
 */
function unparseMetaValue(node) {
    const type = stencila.type(node);
    switch (type) {
        case 'null':
            return {
                t: 'MetaString',
                c: '!!null'
            };
        case 'boolean':
            return {
                t: 'MetaBool',
                c: node
            };
        case 'number':
            return {
                t: 'MetaString',
                c: `!!number ${node}`
            };
        case 'string':
            return {
                t: 'MetaString',
                c: node
            };
        case 'array':
            return {
                t: 'MetaList',
                c: node.map(unparseMetaValue)
            };
        case 'Paragraph':
            return {
                t: 'MetaInlines',
                c: unparseInlines(node.content)
            };
        case 'QuoteBlock':
            return {
                t: 'MetaBlocks',
                c: unparseBlocks(node.content)
            };
        // For both plain objects, and object with `type`, create a `MetaMap`
        case 'object':
        default:
            return {
                t: 'MetaMap',
                c: objectMap(node, (key, value) => unparseMetaValue(value))
            };
    }
}
/**
 * Parse an array of Pandoc `Block` elements.
 */
function parseBlocks(blocks) {
    return blocks.map(block => parseBlock(block));
}
/**
 * Unparse an array of Stencila `BlockContent` nodes.
 */
function unparseBlocks(nodes) {
    return nodes.map(node => unparseBlock(node));
}
/**
 * Parse a Pandoc `Block` element to a Stencila `BlockContent` node.
 */
function parseBlock(block) {
    switch (block.t) {
        case 'Header':
            return parseHeader(block);
        case 'Para':
            return parsePara(block);
        case 'BlockQuote':
            return parseBlockQuote(block);
        case 'CodeBlock':
            return parseCodeBlock(block);
        case 'BulletList':
        case 'OrderedList':
            return parseList(block);
        case 'Table':
            return parseTable(block);
        case 'HorizontalRule':
            return parseHorizontalRule(block);
    }
    throw new Error(`Unhandled Pandoc node type "${block.t}"`);
}
/**
 * Unparse a Stencila `BlockContent` node to a Pandoc `Block` element.
 */
function unparseBlock(block) {
    const type = block.type;
    switch (type) {
        case 'Heading':
            return unparseHeading(block);
        case 'Paragraph':
            return unparseParagraph(block);
        case 'QuoteBlock':
            return unparseQuoteBlock(block);
        case 'CodeBlock':
            return unparseCodeBlock(block);
        case 'List':
            return unparseList(block);
        case 'Table':
            return unparseTable(block);
        case 'ThematicBreak':
            return unparseThematicBreak(block);
    }
    throw new Error(`Unhandled Stencila node type "${block.type}"`);
}
/**
 * Parse a Pandoc `Header` to a Stencila `Heading`
 *
 * Note: currently, any header attributes are ignored
 */
function parseHeader(node) {
    return {
        type: 'Heading',
        depth: node.c[0],
        content: parseInlines(node.c[2])
    };
}
function unparseHeading(node) {
    return {
        t: 'Header',
        c: [node.depth, exports.emptyAttrs, unparseInlines(node.content)]
    };
}
function parsePara(node) {
    return {
        type: 'Paragraph',
        content: parseInlines(node.c)
    };
}
function unparseParagraph(node) {
    return {
        t: 'Para',
        c: unparseInlines(node.content)
    };
}
/**
 * Parse a Pandoc `BlockQuote` to a Stencila `QuoteBlock`.
 */
function parseBlockQuote(node) {
    return {
        type: 'QuoteBlock',
        content: parseBlocks(node.c)
    };
}
/**
 * Unparse a Stencila `QuoteBlock` to a Pandoc `BlockQuote`.
 */
function unparseQuoteBlock(node) {
    return {
        t: 'BlockQuote',
        c: unparseBlocks(node.content)
    };
}
/**
 * Parse a Pandoc `CodeBlock` to a Stencila `CodeBlock`.
 */
function parseCodeBlock(node) {
    const codeblock = {
        type: 'CodeBlock',
        value: node.c[1]
    };
    const attrs = parseAttrs(node.c[0]);
    const language = attrs.classes ? attrs.classes.split(' ')[0] : null;
    if (language)
        codeblock.language = language;
    return codeblock;
}
/**
 * Unparse a Stencila `CodeBlock` to a Pandoc `CodeBlock`.
 */
function unparseCodeBlock(node) {
    const attrs = unparseAttrs({ classes: node.language || '' });
    return {
        t: 'CodeBlock',
        c: [attrs, node.value]
    };
}
/**
 * Parse a Pandoc `BulletList` or `OrderedList` to a Stencila `List`.
 */
function parseList(node) {
    const order = node.t === 'BulletList' ? 'unordered' : 'ascending';
    const blocks = node.t === 'BulletList' ? node.c : node.c[1];
    // TODO: better handling of unwrapping
    const items = blocks.map(blocks => parseBlocks(blocks)[0]);
    return {
        type: 'List',
        order,
        items
    };
}
/**
 * Unparse Stencila `List` as a Pandoc `BulletList` or `OrderedList`.
 */
function unparseList(node) {
    const listAttrs = [
        1,
        { t: Pandoc.ListNumberStyle.DefaultStyle },
        { t: Pandoc.ListNumberDelim.DefaultDelim }
    ];
    const blocks = node.items.map(node => {
        // TODO: need to wrap inline elements if necessary
        return [unparseBlock(node)];
    });
    if (node.order === 'ascending') {
        return { t: 'OrderedList', c: [listAttrs, blocks] };
    }
    else {
        return { t: 'BulletList', c: blocks };
    }
}
/**
 * Parse a Pandoc `Table` to a Stencila `Table`
 *
 * Note: table caption and column widths and alignments
 * are currently ignored.
 */
function parseTable(node) {
    const caption = parseInlines(node.c[0]);
    const aligns = node.c[1];
    const widths = node.c[2];
    const head = node.c[3].map(parseBlocks);
    const data = node.c[4].map(row => row.map(parseBlocks));
    const rows = [head, ...data].map((row) => {
        return {
            type: 'TableRow',
            cells: row.map((cell) => {
                return {
                    type: 'TableCell',
                    // TODO: currently assuming that only one block in each table cell
                    // with `content` property
                    content: cell[0].content
                };
            })
        };
    });
    return {
        type: 'Table',
        rows
    };
}
/**
 * Unparse Stencila `Table` to a Pandoc `Table`.
 */
function unparseTable(node) {
    const caption = [];
    const aligns = [];
    const widths = [];
    let head = [];
    if (node.rows.length > 0) {
        head = node.rows[0].cells.map(cell => {
            // TODO: currently need to wrap stencila.InlineContent[] to pandoc.Block[][]; this will change
            return [
                unparseParagraph({
                    type: 'Paragraph',
                    content: cell.content
                })
            ];
        });
    }
    let rows = [];
    if (node.rows.length > 1) {
        rows = node.rows.slice(1).map((row) => {
            return row.cells.map((cell) => {
                // TODO: ditto todo item above
                return [
                    unparseParagraph({
                        type: 'Paragraph',
                        content: cell.content
                    })
                ];
            });
        });
    }
    return {
        t: 'Table',
        c: [caption, aligns, widths, head, rows]
    };
}
/**
 * Parse a Pandoc `HorizontalRule` to a Stencila `ThematicBreak`
 */
function parseHorizontalRule(node) {
    return {
        type: 'ThematicBreak'
    };
}
/**
 * Unparse a Stencila `ThematicBreak` to a Pandoc `HorizontalRule`
 */
function unparseThematicBreak(node) {
    return {
        t: 'HorizontalRule',
        c: undefined
    };
}
/**
 * Parse an array of Pandoc `Inline` nodes to Stencila `InlineContent` nodes.
 *
 * Merges contiguous `Str` and `Space` elements prior to parsing.
 */
function parseInlines(nodes) {
    const inlines = [];
    let previous;
    for (const node of nodes) {
        if (previous &&
            previous.t === 'Str' &&
            (node.t === 'Space' || node.t === 'Str')) {
            if (node.t === 'Space')
                previous.c += ' ';
            else if (node.t === 'Str')
                previous.c += node.c;
        }
        else {
            inlines.push(node);
            previous = node;
        }
    }
    return inlines.map(parseInline);
}
/**
 * Unparse an array of Stencila `InlineContent` nodes to Pandoc `Inline` nodes.
 */
function unparseInlines(nodes) {
    return nodes.map(unparseInline);
}
/**
 * Parse a Pandoc `Inline` node to a Stencila `InlineContent` node
 */
function parseInline(node) {
    switch (node.t) {
        case 'Space':
            return parseSpace(node);
        case 'Str':
            return parseStr(node);
        case 'Emph':
            return parseEmph(node);
        case 'Strong':
            return parseStrong(node);
        case 'Strikeout':
            return parseStrikeout(node);
        case 'Quoted':
            return parseQuoted(node);
        case 'Code':
            return parseCode(node);
        case 'Link':
            return parseLink(node);
        case 'Image':
            const image = parseImage(node);
            // If the image is an rPNG then decode it and return
            // the embedded node
            const url = image.contentUrl;
            if (url) {
                // TODO: currently assume url is local file, should we fetch remotes?
                const node = rpng.sniffParseSync(url);
                // TODO: avoid `as`
                if (typeof node !== 'undefined')
                    return node;
            }
            return image;
    }
    throw new Error(`Unhandled Pandoc element type "${node.t}"`);
}
function unparseInline(node) {
    const type = stencila.type(node);
    switch (type) {
        case 'string':
            return unparseString(node);
        case 'Emphasis':
            return unparseEmph(node);
        case 'Strong':
            return unparseStrong(node);
        case 'Delete':
            return unparseDelete(node);
        case 'Quote':
            return unparseQuote(node);
        case 'Code':
            return unparseCode(node);
        case 'Link':
            return unparseLink(node);
        case 'ImageObject':
            return unparseImageObject(node);
    }
    return unparseDefault(node);
}
/**
 * Parse a Pandoc `Inline` node to a `string`.
 *
 * For some nodes the Stencila schema requires a `string` whereas
 * the Pandoc AST provides us with an array of inline nodes. An
 * example is the `alt` text for an image. This function is thus
 * intentionally lossy because it will ignore the fact that the
 * text is, for example, emphasized and just return it's string content.
 *
 * For a Pandoc `Note` (e.g. a footnote), the content of the note
 * is ignored. Pandoc `SoftBreak` and `LineBreak` nodes are also
 * parsed to an empty string. For all other nodes, the string
 * content of the node should be returned.
 */
function parseInlineToString(node) {
    switch (node.t) {
        case 'SoftBreak':
        case 'LineBreak':
        case 'Note':
            return '';
        case 'Space':
            return ' ';
        case 'Str':
            return node.c;
        case 'Code':
        case 'Math':
        case 'RawInline':
            return node.c[1];
        case 'Emph':
        case 'Strong':
        case 'Strikeout':
        case 'Superscript':
        case 'Subscript':
        case 'SmallCaps':
            return parseInlinesToString(node.c);
        case 'Quoted':
        case 'Cite':
        case 'Link':
        case 'Image':
        case 'Span':
            return parseInlinesToString(node.c[1]);
    }
}
/**
 * Parse an array of Pandoc `Inline` nodes to a `string`.
 */
function parseInlinesToString(nodes) {
    return nodes.map(parseInlineToString).join('');
}
/**
 * Parse a Pandoc `Space` to a `string`.
 */
function parseSpace(node) {
    return ' ';
}
/**
 * Parse a Pandoc pace` to a `string`.
 */
function parseStr(node) {
    return node.c;
}
/**
 * Unparse a `string` to a Pandoc `Str`.
 */
function unparseString(node) {
    return {
        t: 'Str',
        c: node
    };
}
/**
 * Parse a Pandoc `Emph` to a Stencila `Emphasis`.
 */
function parseEmph(node) {
    return {
        type: 'Emphasis',
        content: parseInlines(node.c)
    };
}
/**
 * Unparse a Stencila `Emphasis` to a Pandoc `Emph`.
 */
function unparseEmph(node) {
    return {
        t: 'Emph',
        c: unparseInlines(node.content)
    };
}
/**
 * Parse a Pandoc `Strong` to a Stencila `Strong`.
 */
function parseStrong(node) {
    return {
        type: 'Strong',
        content: parseInlines(node.c)
    };
}
/**
 * Unparse a Stencila `Strong` to a Pandoc `Strong`.
 */
function unparseStrong(node) {
    return {
        t: 'Strong',
        c: unparseInlines(node.content)
    };
}
/**
 * Parse a Pandoc `Strikeout` to a Stencila `Delete`.
 */
function parseStrikeout(node) {
    return {
        type: 'Delete',
        content: parseInlines(node.c)
    };
}
/**
 * Unparse a Stencila `Delete` to a Pandoc `Strikeout`.
 */
function unparseDelete(node) {
    return {
        t: 'Strikeout',
        c: unparseInlines(node.content)
    };
}
/**
 * Parse a Pandoc `Quoted` to a Stencila `Quote`.
 *
 * Note: the type of quote, single or double, is ignored.
 */
function parseQuoted(node) {
    return {
        type: 'Quote',
        content: parseInlines(node.c[1])
    };
}
/**
 * Unparse a Stencila `Quote` to a Pandoc `Quoted`.
 */
function unparseQuote(node) {
    return {
        t: 'Quoted',
        c: [{ t: Pandoc.QuoteType.SingleQuote }, unparseInlines(node.content)]
    };
}
/**
 * Parse a Pandoc `Code` to a Stencila `Code`.
 */
function parseCode(node) {
    const code = {
        type: 'Code',
        value: node.c[1]
    };
    const attrs = parseAttrs(node.c[0]);
    const language = attrs.classes ? attrs.classes.split(' ')[0] : null;
    if (language)
        code.language = language;
    return code;
}
/**
 * Unparse a Stencila `Code` to a Pandoc `Code`.
 */
function unparseCode(node) {
    const attrs = unparseAttrs({ classes: node.language || '' });
    return {
        t: 'Code',
        c: [attrs, node.value]
    };
}
/**
 * Parse a Pandoc `Link` to a Stencila `Link`.
 *
 * Note: attributes are ignored
 */
function parseLink(node) {
    const [url, title] = node.c[2];
    return {
        type: 'Link',
        content: parseInlines(node.c[1]),
        target: url,
        description: title
    };
}
/**
 * Unparse a Stencila `Link` to a Pandoc `Link`.
 */
function unparseLink(node) {
    const [url, title] = [node.target, node.description || ''];
    return {
        t: 'Link',
        c: [exports.emptyAttrs, unparseInlines(node.content), [url, title]]
    };
}
/**
 * Parse a Pandoc `Image` to a Stencila `ImageObject`.
 *
 * Note: attributes are ignored.
 */
function parseImage(image) {
    const alt = parseInlinesToString(image.c[1]);
    const [url, title] = image.c[2];
    return {
        type: 'ImageObject',
        contentUrl: url,
        title: title,
        text: alt
    };
}
/**
 * Unparse a Stencila `ImageObject` to a Pandoc `Image`.
 */
function unparseImageObject(imageObject) {
    const url = imageObject.contentUrl || '';
    const title = imageObject.title || '';
    const alt = [];
    if (imageObject.text)
        alt.push(unparseString(imageObject.text));
    return {
        t: 'Image',
        c: [exports.emptyAttrs, alt, [url, title]]
    };
}
function unparseDefault(node) {
    const type = stencila.type(node);
    const index = 1;
    const imagePath = `${type.toLowerCase()}-${index}.png`;
    const promise = (async () => {
        const file = await rpng.unparse(node);
        await vfile_1.write(file, imagePath);
    })();
    unparsePromises.push(promise);
    const url = imagePath;
    const title = type;
    return {
        t: 'Image',
        c: [exports.emptyAttrs, [], [url, title]]
    };
}
/**
 * Empty Pandoc element attributes
 */
exports.emptyAttrs = ['', [], []];
/**
 * Parse Pandoc `Attr` attributes to an object
 */
function parseAttrs(node) {
    const attrs = {};
    if (node[0])
        attrs.id = node[0];
    if (node[1])
        attrs.classes = node[1].join(' ');
    for (const attr of node[2])
        attrs[attr[0]] = attr[1];
    return attrs;
}
/**
 * Unparse an object of attributes to a Pandoc `Attr`.
 */
function unparseAttrs(attrs = {}) {
    const { id, classes } = attrs, rest = __rest(attrs, ["id", "classes"]);
    return [id || '', classes ? classes.split(' ') : [], Object.entries(rest)];
}
/**
 * `Array.map` but for `objects`.
 */
function objectMap(obj, func) {
    return Object.assign({}, ...Object.entries(obj).map(([k, v]) => ({ [k]: func(k, v) })));
}
