/**
 * A compiler for Reproducible PNGs (rPNG) files.
 *
 * This compiler parses from, and unparses to, a rPNG which embeds the Stencila node
 * into the `tEXt` chunk of the PNG.
 *
 * This has been implemented here to make use of the HTML converter to
 * render the results. It currently using Puppetter and so will not work
 * in the browser. In the future we may use `html2canvas` and `canvas2image` to enable
 * rPNGs to be [generated in the browser](https://medium.com/@danielsternlicht/capturing-dom-elements-screenshots-server-side-vs-client-side-approaches-6901c706c56f).
 *
 * @see http://www.libpng.org/pub/png/spec/1.2/PNG-Chunks.html#C.Anc-text
 */
/// <reference types="node" />
import * as stencila from '@stencila/schema';
import { Chunk } from 'png-chunks-extract';
import { VFile } from './vfile';
export declare const mediaTypes: string[];
export declare const extNames: string[];
/**
 * Find a text chunk in an image
 *
 * @param keyword The keyword for the text chunk
 * @param chunks The image chunks to search through
 */
export declare function find(keyword: string, chunks: Array<Chunk>): [number, string | undefined];
/**
 * Does an image have a text chunk with the given keyword?
 *
 * @param keyword The keyword for the text chunk
 * @param image The image `Buffer`
 */
export declare function has(keyword: string, image: Buffer): boolean;
/**
 * Extract a text chunk from an image
 *
 * @param keyword The keyword for the text chunk
 * @param image The image `Buffer`
 */
export declare function extract(keyword: string, image: Buffer): string;
/**
 * Insert a text chunk into an image
 *
 * @param keyword The keyword for the text chunk
 * @param text The text to insert
 * @param image The image to insert into
 */
export declare function insert(keyword: string, text: string, image: Buffer): Buffer;
/**
 * Sniff a PNG file to see if it is an rPNG
 *
 * @param content The content to sniff (a file path)
 */
export declare function sniff(content: string): Promise<boolean>;
/**
 * Synchronous version of `sniff()`.
 *
 * @see sniff
 *
 * @param content The content to sniff (a file path)
 */
export declare function sniffSync(content: string): boolean;
/**
 * Parse a rPNG to a Stencila node.
 *
 * This is done by extracting the JSON
 * from the `tEXt` chunk and parsing it.
 *
 * @param file The `VFile` to parse
 * @returns The Stencila node
 */
export declare function parse(file: VFile): Promise<stencila.Node>;
/**
 * Synchronous version of `parse()`.
 *
 * @see parse
 *
 * @param content The content to sniff (a file path).
 */
export declare function parseSync(file: VFile): stencila.Node;
/**
 * Sniff and parse a file if it is a rPNG.
 *
 * This function is like combining `sniffSync()` and `parseSync()`
 * but is faster because it only reads the file contents once.
 *
 * @param filePath The file path to sniff.
 */
export declare function sniffParseSync(filePath: string): stencila.Node | undefined;
/**
 * Unparse a Stencila node to a rPNG.
 *
 * This is done by dumping the node to HTML,
 * "screen-shotting" the HTML to a PNG and then inserting the
 * node's JSON into the image's `tEXt` chunk.
 *
 * @param node The Stencila node to unparse
 * @param filePath The file system path to write to
 */
export declare function unparse(node: stencila.Node, filePath?: string): Promise<VFile>;
