"use strict";
/**
 * A compiler for Reproducible PNGs (rPNG) files.
 *
 * This compiler parses from, and unparses to, a rPNG which embeds the Stencila node
 * into the `tEXt` chunk of the PNG.
 *
 * This has been implemented here to make use of the HTML converter to
 * render the results. It currently using Puppetter and so will not work
 * in the browser. In the future we may use `html2canvas` and `canvas2image` to enable
 * rPNGs to be [generated in the browser](https://medium.com/@danielsternlicht/capturing-dom-elements-screenshots-server-side-vs-client-side-approaches-6901c706c56f).
 *
 * @see http://www.libpng.org/pub/png/spec/1.2/PNG-Chunks.html#C.Anc-text
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_extra_1 = __importDefault(require("fs-extra"));
const path_1 = __importDefault(require("path"));
const png_chunk_text_1 = __importDefault(require("png-chunk-text"));
const png_chunks_encode_1 = __importDefault(require("png-chunks-encode"));
const png_chunks_extract_1 = __importDefault(require("png-chunks-extract"));
const punycode_1 = __importDefault(require("punycode"));
const puppeteer_1 = __importDefault(require("puppeteer"));
const boot_1 = require("./boot");
const index_1 = require("./index");
const stencila_css_template_1 = require("./templates/stencila-css-template");
const vfile_1 = require("./vfile");
// A vendor media type similar to https://www.iana.org/assignments/media-types/image/vnd.mozilla.apng
// an custom extension to be able to refere to this format more easily.
exports.mediaTypes = ['vnd.stencila.rpng'];
exports.extNames = ['rpng'];
/**
 * The keyword to use for the PNG chunk containing the JSON
 */
const KEYWORD = 'JSON';
/**
 * Find a text chunk in an image
 *
 * @param keyword The keyword for the text chunk
 * @param chunks The image chunks to search through
 */
function find(keyword, chunks) {
    let index = 0;
    for (let chunk of chunks) {
        if (chunk.name === 'tEXt') {
            const entry = png_chunk_text_1.default.decode(chunk.data);
            if (entry.keyword === keyword) {
                return [index, punycode_1.default.decode(entry.text)];
            }
        }
        index += 1;
    }
    return [-1, undefined];
}
exports.find = find;
/**
 * Does an image have a text chunk with the given keyword?
 *
 * @param keyword The keyword for the text chunk
 * @param image The image `Buffer`
 */
function has(keyword, image) {
    const chunks = png_chunks_extract_1.default(image);
    const [index, text] = find(keyword, chunks);
    return text ? true : false;
}
exports.has = has;
/**
 * Extract a text chunk from an image
 *
 * @param keyword The keyword for the text chunk
 * @param image The image `Buffer`
 */
function extract(keyword, image) {
    const chunks = png_chunks_extract_1.default(image);
    const [index, text] = find(keyword, chunks);
    if (!text)
        throw Error('No chunk found');
    return text;
}
exports.extract = extract;
/**
 * Insert a text chunk into an image
 *
 * @param keyword The keyword for the text chunk
 * @param text The text to insert
 * @param image The image to insert into
 */
function insert(keyword, text, image) {
    const chunks = png_chunks_extract_1.default(image);
    const [index, current] = find(keyword, chunks);
    if (current)
        chunks.splice(index, 1);
    const chunk = png_chunk_text_1.default.encode(keyword, punycode_1.default.encode(text));
    chunks.splice(-1, 0, chunk);
    return Buffer.from(png_chunks_encode_1.default(chunks));
}
exports.insert = insert;
/**
 * Sniff a PNG file to see if it is an rPNG
 *
 * @param content The content to sniff (a file path)
 */
async function sniff(content) {
    if (path_1.default.extname(content) === '.png') {
        if (await fs_extra_1.default.pathExists(content)) {
            const contents = await fs_extra_1.default.readFile(content);
            return has(KEYWORD, contents);
        }
    }
    return false;
}
exports.sniff = sniff;
/**
 * Synchronous version of `sniff()`.
 *
 * @see sniff
 *
 * @param content The content to sniff (a file path)
 */
function sniffSync(content) {
    if (path_1.default.extname(content) === '.png') {
        if (fs_extra_1.default.existsSync(content)) {
            const contents = fs_extra_1.default.readFileSync(content);
            return has(KEYWORD, contents);
        }
    }
    return false;
}
exports.sniffSync = sniffSync;
/**
 * Parse a rPNG to a Stencila node.
 *
 * This is done by extracting the JSON
 * from the `tEXt` chunk and parsing it.
 *
 * @param file The `VFile` to parse
 * @returns The Stencila node
 */
async function parse(file) {
    return parseSync(file);
}
exports.parse = parse;
/**
 * Synchronous version of `parse()`.
 *
 * @see parse
 *
 * @param content The content to sniff (a file path).
 */
function parseSync(file) {
    if (Buffer.isBuffer(file.contents)) {
        const json = extract(KEYWORD, file.contents);
        return JSON.parse(json);
    }
    return {};
}
exports.parseSync = parseSync;
/**
 * Sniff and parse a file if it is a rPNG.
 *
 * This function is like combining `sniffSync()` and `parseSync()`
 * but is faster because it only reads the file contents once.
 *
 * @param filePath The file path to sniff.
 */
function sniffParseSync(filePath) {
    if (path_1.default.extname(filePath) === '.png') {
        if (fs_extra_1.default.existsSync(filePath)) {
            const image = fs_extra_1.default.readFileSync(filePath);
            const chunks = png_chunks_extract_1.default(image);
            const [index, json] = find(KEYWORD, chunks);
            if (json)
                return JSON.parse(json);
        }
    }
}
exports.sniffParseSync = sniffParseSync;
/**
 * Unparse a Stencila node to a rPNG.
 *
 * This is done by dumping the node to HTML,
 * "screen-shotting" the HTML to a PNG and then inserting the
 * node's JSON into the image's `tEXt` chunk.
 *
 * @param node The Stencila node to unparse
 * @param filePath The file system path to write to
 */
async function unparse(node, filePath) {
    // Generate HTML of the 'value' of the node, which depends on the
    // node type. In the future, we may make this part of the schema definitions
    // and have a `stencila.value()` function to retrieve the value for the node
    // But currently just using this...
    let value;
    if (node && typeof node === 'object' && node.hasOwnProperty('value')) {
        value = node.value;
    }
    else {
        value = node;
    }
    let html = await index_1.dump(value, 'html');
    // Generate image of rendered HTML
    const browser = await puppeteer_1.default.launch({
        executablePath: boot_1.chromiumPath
    });
    const page = await browser.newPage();
    await page.addStyleTag({ content: stencila_css_template_1.stencilaCSS });
    await page.addScriptTag({
        url: 'https://cdnjs.cloudflare.com/ajax/libs/highlight.js/9.15.6/highlight.min.js'
    });
    await page.setContent(`<div id="target" style="display: inline-block; padding: 0.1rem">${html}</div>`, { waitUntil: 'networkidle0' });
    const elem = await page.$('#target');
    if (!elem)
        throw new Error('Element not found!');
    // Run Highlight.js on any found code blocks
    await page.evaluate('document.querySelectorAll("pre code").forEach(block => hljs.highlightBlock(block))');
    const buffer = await elem.screenshot({
        encoding: 'binary'
    });
    await browser.close();
    // Insert JSON of the thing into the image
    const json = JSON.stringify(node);
    const image = insert(KEYWORD, json, buffer);
    const file = vfile_1.load(image);
    if (filePath) {
        file.path = filePath;
        await vfile_1.write(file, filePath);
    }
    return file;
}
exports.unparse = unparse;
