"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const ajv_1 = __importDefault(require("ajv"));
const better_ajv_errors_1 = __importDefault(require("better-ajv-errors"));
const fs_extra_1 = __importDefault(require("fs-extra"));
const globby_1 = __importDefault(require("globby"));
const immer_1 = __importDefault(require("immer"));
const path_1 = __importDefault(require("path"));
const person_1 = require("./person");
const built = path_1.default.join(path_1.default.dirname(require.resolve('@stencila/schema')), 'built');
// Load all schemas for use in by Ajv validator
const schemas = globby_1.default
    .sync(path_1.default.join(built, '*.schema.json'))
    .map(file => fs_extra_1.default.readJSONSync(file));
// Read in aliases for use in coerce function
const aliases = fs_extra_1.default.readJSONSync(path_1.default.join(built, 'aliases.json'));
/**
 * Create a node of a type
 * @param type The name of the type
 * @param initial Initial values for properties
 * @param validation What validation should done?
 */
function create(type, initial = {}, validation = 'validate') {
    let node = Object.assign({ type }, initial);
    if (validation === 'validate')
        return validate(node, type);
    else if (validation === 'coerce')
        return coerce(node, type);
    else
        return node;
}
exports.create = create;
/**
 * Get the type of a node
 * @param node The node to get the type for
 */
function type(node) {
    if (node === null)
        return 'null';
    let type = typeof node;
    if (type === 'object') {
        if (Array.isArray(node))
            return 'array';
        if (node.type)
            return node.type;
    }
    return type;
}
exports.type = type;
/**
 * Is a node of a particular type/s
 * @param node The node to check
 * @param types The type names to check against
 */
function is(node, types) {
    if (typeof types === 'string')
        return type(node) === types;
    else
        return types.includes(type(node));
}
exports.is = is;
/**
 * Assert that a node is of a particular type/s
 * @param node The node to check
 * @param types The type names to check against
 */
function assert(node, types) {
    if (is(node, types))
        return true;
    else {
        const list = typeof types === 'string' ? types : types.join('|');
        throw new Error(`Node type is "${type(node)}" but expected "${list}"`);
    }
}
exports.assert = assert;
/**
 * Cast a node to a particular type
 *
 * The node is validated against the type.
 * This means that an error will be throw if during:
 *   - up-casting the node does not have properties
 *     that are required by the schema of the new type
 *   - down-casting the node has properties that
 *     are additional to those in the schema of the new type
 * Use `coerce` if you want to ignore such errors
 * and force mutating the node to the type.
 *
 * @param node The node to cast
 * @param type The type to cast to
 */
function cast(node, type) {
    return immer_1.default(node, (casted) => {
        casted.type = type;
        validate(casted, type);
    });
}
exports.cast = cast;
// Cached JSON Schema validation functions
const validators = new ajv_1.default({
    schemas,
    jsonPointers: true
});
/**
 * Validate a node against a type's schema
 * @param node The node to validate
 * @param type The type to validate against
 */
function validate(node, type) {
    const validator = validators.getSchema(`https://stencila.github.com/schema/${type}.schema.json`);
    if (!validator)
        throw new Error(`No schema for type "${type}".`);
    if (!validator(node)) {
        const errors = better_ajv_errors_1.default(validator.schema, node, validator.errors, {
            format: 'js'
        });
        throw new Error(errors.map(error => `${error.error}`).join(';'));
    }
    return node;
}
exports.validate = validate;
/**
 * Is a node valid with respect to a particular type's schema
 * @param node The node to check
 * @param type The type to check against
 */
function valid(node, type) {
    try {
        validate(node, type);
        return true;
    }
    catch (error) {
        return false;
    }
}
exports.valid = valid;
// Cached JSON Schema validation/mutation functions
// These use Ajv options that coerce nodes so we
// keep them separate from pure non-mutating validators.
const mutators = new ajv_1.default({
    schemas,
    jsonPointers: true,
    // Add values from `default` keyword when property is missing
    useDefaults: true,
    // Remove any additional properties
    removeAdditional: true,
    // Coerce type of data to match type keyword and coerce scalar
    // data to an array with one element and vice versa, as needed.
    coerceTypes: 'array'
});
/**
 * A list of parsers that can be applied using the `parser` keyword
 */
const parsers = {
    csv,
    ssv,
    person: person_1.parse
};
/**
 * Parse comma separated string data into an array of strings
 */
function csv(data) {
    return data.split(',');
}
/**
 * Parse space separated string data into an array of strings
 */
function ssv(data) {
    return data.split(/ +/);
}
/**
 * Custom validation function that handles the `parser`
 * keyword.
 */
const parserValidate = (parser, data, parentSchema, dataPath, parentData, parentDataProperty, rootData) => {
    function raise(msg) {
        parserValidate.errors = [
            {
                keyword: 'parser',
                dataPath: '' + dataPath,
                schemaPath: '',
                params: {
                    keyword: 'parser'
                },
                message: msg,
                data: data
            }
        ];
        return false;
    }
    const parse = parsers[parser];
    if (!parse)
        return raise(`no such parser: "${parser}"`);
    let parsed;
    try {
        parsed = parse(data);
    }
    catch (error) {
        const parseError = error.message.split('\n')[0];
        return raise(`error when parsing using "${parser}": ${parseError}`);
    }
    if (parentData !== undefined && parentDataProperty !== undefined) {
        ;
        parentData[parentDataProperty] = parsed;
    }
    return true;
};
mutators.addKeyword('parser', {
    type: 'string',
    modifying: true,
    validate: parserValidate
});
/**
 * Coerce a node so it conforms to a type's schema
 *
 * @param node The node to coerce
 * @param type The type to conform to
 */
function coerce(node, type) {
    const mutator = mutators.getSchema(`https://stencila.github.com/schema/${type}.schema.json`);
    if (!mutator)
        throw new Error(`No schema for type "${type}".`);
    return immer_1.default(node, (coerced) => {
        if (typeof coerced === 'object')
            coerced.type = type;
        // Rename property aliases
        rename(coerced);
        // coerce and validate
        if (!mutator(coerced)) {
            const errors = better_ajv_errors_1.default(mutator.schema, node, mutator.errors, {
                format: 'js'
            });
            throw new Error(errors.map(error => `${error.error}`).join(';'));
        }
    });
    // Replace aliases with canonical names
    function rename(node) {
        if (!node || typeof node !== 'object')
            return;
        if (!(node.type && aliases[node.type]))
            return;
        let propertyAliases = aliases[node.type];
        for (let [key, child] of Object.entries(node)) {
            if (!Array.isArray(node)) {
                const name = propertyAliases[key];
                if (name) {
                    node[name] = child;
                    delete node[key];
                }
            }
            rename(child);
        }
    }
}
exports.coerce = coerce;
