"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_extra_1 = __importDefault(require("fs-extra"));
const get_stdin_1 = __importDefault(require("get-stdin"));
// @ts-ignore
const to_vfile_1 = __importDefault(require("to-vfile"));
const vfile_1 = __importDefault(require("vfile"));
/**
 * Create a virtual file
 *
 * @param contents Raw string contents of `VFile`, or file path
 * @param options Options to `vfile` see https://github.com/vfile/vfile#vfileoptions
 */
function create(contents, options = {}) {
    if (contents) {
        if (isPath(contents))
            options = Object.assign({}, options, { path: contents });
        else
            options = Object.assign({}, options, { contents });
    }
    return vfile_1.default(options);
}
exports.create = create;
/**
 * Load a string into a virtual file
 *
 * @param contents Contents to load
 */
function load(contents) {
    return vfile_1.default({ contents });
}
exports.load = load;
/**
 * Dump a string from the contents of a virtual files
 *
 * If the file has no `contents` but does have a file
 * `path`, then `read` the virtual file.
 *
 * @param file The virtual file to dump
 */
async function dump(vfile) {
    if (vfile.contents)
        return vfile.toString();
    if (vfile.path) {
        const readFile = await to_vfile_1.default.read(vfile.path);
        return readFile.toString();
    }
    return '';
}
exports.dump = dump;
/**
 * Read a file path into a virtual file
 *
 * @param path The content, or file system path to read from. If `-`, read from stdin.
 */
async function read(content) {
    if (!content)
        throw new Error('Argument `path` is required');
    if (content === '-') {
        return load(await get_stdin_1.default());
    }
    else if (isPath(content)) {
        return to_vfile_1.default.read(content);
    }
    else {
        return load(content);
    }
}
exports.read = read;
/**
 * Write a virtual file to the file system or standard output.
 *
 * Will never write an empty file i.e. will only write if the
 * `vfile` has `contents`.
 *
 * @param vfile Virtual file to write
 * @param path The file system path to write to, or `-` to write to stdout.
 */
async function write(vfile, path) {
    if (!path)
        throw new Error('Argument `path` is required');
    if (path === '-') {
        console.log(await dump(vfile));
    }
    else if (path && vfile.contents) {
        vfile.path = path;
        await to_vfile_1.default.write(vfile);
    }
}
exports.write = write;
/**
 * Is a string a file system path?
 *
 * Several functions in this module allow for content to be passed as a
 * string of raw content (e.g. `This is some Markdown`) or as a file system
 * path. This function, assesses whether a file is a file system path or not.
 *
 * If the string starts with `/`, `./`, `../` (or Windows compatible backslash
 * equivalents as well as drive letter prefixed strings e.g. `C:\`)
 * then it is assumed to be a path but the presence of the path is not checked.
 * The reason for not checking presence
 * here is because "if the content looks like a path then the user probably meant
 * it to be a path". That is, if a user tries to convert the string "./file.md" then,
 * if that file doesn't exist, they probably want the Markdown converter to give me an error
 * message. They probably don't want the `match` function to try and find some other converter
 * that acts on a plain strings.
 *
 * For all other strings, this function does check for presence, returning `true`
 * if the file exists.
 *
 * @param content The string to assess.
 */
function isPath(content) {
    if (/^(\/)|(\\)|([A-Z]:\\)|(\.(\/|\\))|(\.\.(\/|\\))/.test(content)) {
        return true;
    }
    if (fs_extra_1.default.existsSync(content))
        return true;
    return false;
}
exports.isPath = isPath;
